# Holiday Utils

A library for efficiently defining and generating recurring holidays.

## Dependencies

Requires `python-dateutil`.  To support additional calendar sets, the extras may
be:

- `hebrew`: `pyluach`
- `islamic`: `hijriconverter`
- `lunar`: `LunarCalendar`

To install all dependencies, try:

`pip install "holiday-utils[hebrew, islamic, lunar]"`

## Defining Holidays

The library is built around the idea of a `HolidayGenerator`. These classes
know how to dynamically generate a sequence of `datetime.date` instances based
on a start date, and a possible end date.  If no end date is given, the
generator may continue to produce dates forever.

The `HolidayGenerator` interface includes one public method:
`dates(after: date, before: date = None)`. There are a number of base classes
to assist in making an efficient holiday generator.  The dates generated should
include the `after` and (if given) `before` dates in their potential set of
responses

## `HolidayGenerator`s

### `PredefinedHolidayGenerator`

It may be that the occurrence of a holiday is not easy or feasible to calculate
dynamically, and generating dates ahead of time is easiest.  This class takes
a list of dates, and will dispense them as needed.

```
hard_to_calc = PredefinedHolidayGenerator([date(2019, 3, 4), date(2020, 9, 3)])
```

### `StableDateGenerator`

If a holiday occurs every year on the Gregorian calendar at the same month and
day, a StableDateGenerator is all that's necessary.

```
new_years_day = StableDateGenerator(1, 1)
christmas = StableDateGenerator(12, 25)
```

### `FloatingDateGenerator`

If a holiday occurs once per Gregorian year, but the date may float around,
subclassing FloatingDateGenerator and overriding `date_for_year(year: int)`
should return a single `datetime.date` instance, or None, if no such event
occurs in the given year.

```
class FakeLeapDayHoliday(FloatingDateGenerator):

    def date_for_year(self, year: int) -> Optional[date]:
        try:
            return date(year, 2, 29)
        except ValueError:
            return None
```

### `DateUtilRRuleGenerator`

If a holiday is easily defined by a `dateutil.rrule.RRule` instance, this may
be used to wrap the RRule interface.

```
from dateutil.rrule import rrule, WEEKLY, TH
payday = rrule(freq=WEEKLY, interval=2, byweekday=TH, dtstart=date(2019, 1, 3))

payday_gen = DateUtilRRuleGenerator(payday)
```

### `DateUtilEasterGenerator`

A special case that uses the `dateutil.easter` calculator to generate the day
for Easter.

## `HolidayGenerator`s built with other `HolidayGenerator`s

### `HolidayRelativeGenerator`

If a holiday occurs in relation to an already defined generator, instances will
make a date instance at a given offset.  The constructor takes a
`HolidayGenerator` instance, and an optional `datetime.timedelta` instance.  If
`delta` isn't defined, the default is to generate one day before the holiday.

```
easter = DateUtilEasterGenerator()
good_friday = HolidayRelativeGenerator(easter, timedelta(days=-2))
```

### `NearestWeekdayGenerator`

If a holiday has meaning for business day downtime, wherein every year its
observance may be shifted by one day so that workers may take a holiday, this
will move the generated date either one day forward or backward, as necessary
to move the date to a weekday.

```
fourth_of_july = StableDateGenerator(7, 4)
fourth_of_july_obsv = NearestWeekdayGenerator(fourth_of_july)
```

## `HolidayCollection`

It is usually a set of holidays that are being tested for. A
`HolidayCollection` wraps a collection of `HolidayGenerator` and will in turn
generate all dates from those generators in the order in which they occur.  The
optional `dedupe` keyword argument indicates whether dates that are generated
multiple times should be generated in succession.

```
my_birthday = StableDateGenerator(1, 1)
holidays = [new_years_day, my_birthday, fourth_of_july]
my_holidays = HolidayCollection(holidays)

list(my_holidays.dates(date(2019, 1, 1), date(2019, 12, 31)))
# [date(2019, 1, 1), date(2019, 7, 4)]

my_holidays = HolidayCollection(holidays, dedupe=False)

list(my_holidays.dates(date(2019, 1, 1), date(2019, 12, 31)))
# [date(2019, 1, 1), date(2019, 1, 1), date(2019, 7, 4)]
```

## Pre-Defined Holidays

### US Holidays

All instances are importable from `holiday_utils.us_national`

- `NewYears`
- `NewYearsEve`
- `MartinLutherKingJrDay`
- `PresidentsDay` (a.k.a. Washington’s Birthday)
- `ValentinesDay`
- `StPatricksDay`
- `AprilFoolsDay`
- `MemorialDay`
- `IndependenceDay`
- `IndependenceDayObserved`
- `LaborDay`
- `IndigenousPeoplesDay` (a.k.a. Columbus Day)
- `Halloween`
- `VeteransDay`
- `VeteransDayObserved`
- `Thanksgiving`
- `BlackFriday` (a.k.a. day after Thanksgiving)
- `ElectionDay`

### US Christian Holidays

importable from `holiday_utils.us_christian`

- `Christmas`
- `ChristmasEve`
- `ChristmasObserved`
- `Easter`
- `GoodFriday`
- `AshWednesday`
- `PalmSunday`

### Hebrew

importable from `holiday_utils.hebrew`.
Note: all of the dates generated are for the Gregorian date containing the
night of the beginnning of the holiday.

- `RoshHashannah`
- `YomKippur`
- `Passover`
- `FirstNightHanukkah`

### Islamic

importable from `holiday_utils.islamic`

- `Ashura`
- `Ramadan`
- `EidAlFatr`
- `EidAlAdha`

### Lunar

importable from `holiday_utils.lunar`

- `LunarNewYear` (a.k.a Chinese New Year)

## Utilities

### `holiday_utils.utils.next_non_holiday(base_date: date, holiday_gen: HolidayGenerator, days_of_week: Container[int] = None) -> date`

Find the next valid date matching the following Criteria:
- the date does not fall on a Holiday (as defined by `holiday_gen`)
- the date is on a weekday as defined in `days_of_week` (if None, any day of the week is valid). Days of the week are integers where 0 = Monday, 6 = Sunday
