from datetime import date, datetime, time, timedelta
from itertools import dropwhile, takewhile
from operator import itemgetter
from typing import Collection, Iterator, Optional, Sequence, cast

try:
    from typing import Literal
except ImportError:  # pragma: no cover
    from typing_extensions import Literal  # type: ignore

ONE_DAY = timedelta(days=1)


class HolidayGenerator:
    """\
    An abstract base class for building generators of repeating dates,
    generally on a yearly schedule, but not necessarily.

    Subclasses must implement `date_gen(after: date)` which is a generator
    of date instances, beginning on `after` (inclusive).
    """

    def date_gen(self, after: date) -> Iterator[date]:
        raise NotImplementedError

    def dates(self, after: date, before: Optional[date] = None) -> Iterator[date]:
        """\
        Returns a generator of dates between after and before (inclusive)
        """
        gen: Iterator[date] = dropwhile(lambda d: d < after, self.date_gen(after))

        if before is not None:
            before = cast(date, before)
            gen = takewhile(lambda d: d <= before, gen)  # type: ignore

        return gen


class PredefinedHolidayGenerator(HolidayGenerator):
    """\
    A HolidayGenerator that takes a list of date instances, for holidays where
    its easier to pre-define a set of dates, rather than build a calculator.
    """

    def __init__(self, dates: Sequence[date]):
        self._dates = dates

    def date_gen(self, after: date) -> Iterator[date]:
        return iter(self._dates)


class FloatingDateGenerator(HolidayGenerator):
    """\
    A HolidayGenerator that assumes that there is a single date during the year
    that can be calculated easily.

    Subclasses must override `date_for_year(year: int)` to provide the date
    that exists that year.
    """

    def date_for_year(self, year: int) -> Optional[date]:
        raise NotImplementedError

    def date_gen(self, after: date) -> Iterator[date]:
        year = after.year

        while True:
            d = self.date_for_year(year)
            if d is None:
                break
            yield d
            year += 1


class StableDateGenerator(FloatingDateGenerator):
    """\
    A HolidayGenerator that occurs on the same month and day of every year.
    """

    def __init__(self, month: int, day: int):
        self.month = month
        self.day = day

    def date_for_year(self, year: int) -> Optional[date]:
        return date(year, self.month, self.day)


class HolidayRelativeGenerator(HolidayGenerator):
    """\
    A HolidayGenerator that builds dates relative to the dates of another
    generator’s output.  This is useful for days such as the day after
    Thanksgiving (with `delta = timedelta(days=1)`), or Good Friday
    (`delta = timedelta(days=-3)`)
    """

    def __init__(self, parent_gen: HolidayGenerator, delta: timedelta = -ONE_DAY):
        self.gen = parent_gen
        self.delta = delta

    def date_gen(self, after: date) -> Iterator[date]:
        if self.delta > timedelta(days=0):
            after = after - self.delta
        for d in self.gen.date_gen(after):
            yield d + self.delta


class NearestWeekdayGenerator(HolidayGenerator):
    """\
    A HolidayGenerator that produces dates to the nearest weekday to a holiday
    produced by another HolidayGenerator.  If `include_original` is `True`,
    the originally generated date will be included in the generated output,
    otherwise just the nearest weekday will be included.  This is useful for
    holidays that are 'observed' in the event the actual date occurs on a
    Saturday or Sunday.
    """

    def __init__(self, parent_gen: HolidayGenerator, *, include_original: bool = False):
        self.gen = parent_gen
        self.include_original = include_original

    def date_gen(self, after: date) -> Iterator[date]:
        for d in self.gen.date_gen(after):
            dates = [d] if self.include_original else []

            if d.weekday() == 5:
                dates.insert(0, d - ONE_DAY)
            elif d.weekday() == 6:
                dates.append(d + ONE_DAY)
            elif not self.include_original:
                dates.append(d)

            yield from dates


class DateUtilRRuleGenerator(HolidayGenerator):
    def __init__(self, rrule):
        self.rrule = rrule

    def date_gen(self, after: date) -> Iterator[date]:
        afterdt = datetime.combine(after, time())
        return (dt.date() for dt in self.rrule.xafter(afterdt, inc=True))


class DateUtilEasterGenerator(FloatingDateGenerator):
    def __init__(self, method: Literal[1, 2, 3] = None):
        self.method = method

    def date_for_year(self, year: int) -> date:
        from dateutil.easter import easter

        if self.method:
            return easter(year, self.method)

        return easter(year)


class HolidayCollection(HolidayGenerator):
    """\
    A HolidayGenerator that will generate from a collection of
    HolidayGenerators.  If `dedupe` is True, instances of dates will not be
    duplicated, even if multiple generators produce the same date.
    """

    def __init__(self, gens: Collection[HolidayGenerator], *, dedupe: bool = True):
        self.gens = gens
        self.dedupe = dedupe

    def date_gen(self, after: date) -> Iterator[date]:
        gens = []

        for gen in self.gens:
            g = gen.dates(after)
            next_date = next(g, None)
            if next_date:
                gens.append((next_date, g))

        last_date = None
        while len(gens):
            gens.sort(key=itemgetter(0))
            min_date, g = gens.pop(0)
            if not self.dedupe or last_date != min_date:
                yield min_date
            last_date = min_date
            next_date = next(g, None)
            if next_date:
                gens.append((next_date, g))
